/*
 *
 * This source file belongs to the Sensoria CASE tool developed at the Institute of 
 * Computer Science, Programming and Software Engineering, of the Ludwig-Maximilians-
 * Universität München, Germany. 
 *
 * See the enclosed LICENSE file for licensing information.
 *
 */
package com.spinroot.spin.ws;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

/**
 * 
 * This class implements the SPIN Web Service.
 * 
 * @version $Id: RunSpin.java,v 1.2 2007/03/09 13:13:07 mayer Exp $
 * @author Max Zeumer
 */
public class RunSpin {

	private List<Integer> fileNumbers= new ArrayList<Integer>();

	private String next= "-------------------------------------------------------------------------------------------------------\n";

	private final String spinLocation= "spin";

	private final String gccLocation= "gcc";

	private final String tmpFolderLocation= "tmp/spin/";

	public static final int PROPERTYCHECK= 1;

	public static final int TRACE= 2;

	public static final int LOG= 0;

	private Hashtable<String, String> commandOptions;

	/**
	 * @param file
	 * @param params
	 * @return
	 */
	public String[] checkPromelaFile(String pmlfilecontent, String pmlltlfilecontent, String[] options) {

		if (pmlltlfilecontent == null)
			pmlltlfilecontent= "";

		// create a unique id to use as a filename
		// to avoid mix ups with parallel requests
		int fileNumber= 1;
		while (fileNumbers.contains(new Double(fileNumber))) {
			fileNumber= (int) Math.random();
		}
		fileNumbers.add(new Integer(fileNumber));

		// set up Strings to return later
		String propertyCheck= "";
		String trace= "";
		String log= "";
		String[] result= new String[3];

		commandOptions= parseCommandOptions(options);

		log+= "[sf_ID_LOG: checkPromelaFile] Command line properties: \n";
		log+= getStringRepresentation(commandOptions);

		String call;

		// create the files in tmp folder
		File pmlFile= new File(tmpFolderLocation + "spin" + new Integer(fileNumber).toString() + ".pml");
		File pmlltlFile= new File(tmpFolderLocation + "spin" + new Integer(fileNumber).toString() + ".pml.ltl");
		File ltlFile= new File(tmpFolderLocation + "spin" + new Integer(fileNumber).toString() + ".ltl");
		try {
			// create the tmp directories and folders
			new File(pmlFile.getParent()).mkdirs();
			pmlFile.createNewFile();
			new File(pmlltlFile.getParent()).mkdirs();
			pmlltlFile.createNewFile();
			new File(ltlFile.getParent()).mkdirs();
			ltlFile.createNewFile();
			// write the parameters to the actual files
			FileWriter pmlFileWriter= new FileWriter(pmlFile);
			FileWriter pmlltlFileWriter= new FileWriter(pmlltlFile);
			pmlFileWriter.write(pmlfilecontent);
			pmlFileWriter.flush();
			if (!pmlltlfilecontent.equals("")) {
				pmlltlFileWriter.write(pmlltlfilecontent);
				pmlltlFileWriter.flush();
			}
		} catch (IOException e) {
			// instantly return error message
			propertyCheck= "ERROR: An unhandled Exception occured while processing your request. See log for details. \n";
			log+= "[ERROR: checkPromelaFile] " + e.getMessage();
			for (int i= 0; i < e.getStackTrace().length; i++) {
				log+= "[JAVA: checkPromelaFile] " + e.getStackTrace()[i].toString() + "\n";
			}
			result[PROPERTYCHECK]= propertyCheck;
			result[TRACE]= trace;
			result[LOG]= log;
			return result;
		}

		if (pmlfilecontent.indexOf("never {") >= 0) {
			try {
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Case: Never claim is included in pml string. \n" + next;

				// running spin
				if (commandOptions.get("claimINpml.spin").equals("")) {
					call= spinLocation + " -a " + pmlFile.getAbsolutePath();
				} else {
					call= spinLocation + " " + commandOptions.get("claimINpml.spin") + " " + pmlFile.getAbsolutePath();
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Verifying model (" + call + ")\n" + next;

				Process process= Runtime.getRuntime().exec(call);
				log= writeOutputToLog(process, log);
				process.waitFor();
				if (process.exitValue() != 0)
					throw new Exception("An error occured while running Spin.\n");

				// compiling pan.c with gcc
				if (commandOptions.get("claimINpml.gcc").equals("")) {
					call= gccLocation + " -DXUSAFE -DNOREDUCE -DCOLLAPSE -o pan pan.c";
				} else {
					call= gccLocation + " " + commandOptions.get("claimINpml.gcc") + " -o pan pan.c";
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Compiling (" + call + ")\n" + next;
				process= Runtime.getRuntime().exec(call);
				log= writeOutputToLog(process, log);
				process.waitFor();

				// execute pan.exe
				String pan= new File("pan").getAbsolutePath();
				if (commandOptions.get("claimINpml.pan").equals("")) {
					call= pan;
				} else {
					call= pan + " " + commandOptions.get("claimINpml.pan");
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Executing (" + call + ")\n" + next;
				process= Runtime.getRuntime().exec(call);
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Output from executing pan: \n" + next;
				log= writeOutputToLog(process, log);
				process.waitFor();

			} catch (Exception e1) {
				// instantly return error message
				propertyCheck= "ERROR: An exception occured while processing your request. See log for details. \n";
				log+= "[ERROR: checkPromelaFile] " + e1.getMessage();
				for (int i= 0; i < e1.getStackTrace().length; i++) {
					log+= "[JAVA: checkPromelaFile] " + e1.getStackTrace()[i].toString() + "\n";
				}
				result[PROPERTYCHECK]= propertyCheck;
				result[TRACE]= trace;
				result[LOG]= log;
				return result;
			}
		} else if (!pmlltlfilecontent.equals("")) {
			try {
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Case: Never claim is NOT included in pml string and pml.ltl string is NOT empty. \n"
						+ next;

				// access .pml.ltl file content
				String line;
				String strConstraint= "";
				RandomAccessFile pmlltlRAF;
				pmlltlRAF= new RandomAccessFile(pmlltlFile, "r");

				// filter the relevant lines
				while ( (line= pmlltlRAF.readLine()) != null) {
					if (! (line.indexOf("#") == 0 || line.equals("")))
						strConstraint+= line + "\n";
				}
				if (strConstraint.equals(""))
					throw new Exception("File does not contain a claim. \n");

				// store the constraint in the .ltl file
				FileWriter ltlFileWriter= new FileWriter(ltlFile);
				ltlFileWriter.write(strConstraint);
				ltlFileWriter.flush();

				// translate ltl into never claim
				String ltl2never= commandOptions.get("claimINpmlltl.ltl2never");
				if (ltl2never == null || ltl2never.equals("")) {
					call= spinLocation + " -F " + ltlFile.getAbsolutePath();
				} else {
					call= spinLocation + " " + ltl2never + " " + ltlFile.getAbsolutePath();
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Translate ltl into never claim (" + call + ")\n" + next;
				Process process= Runtime.getRuntime().exec(call);

				// capture output
				String neverClaim= pmlltlfilecontent + "\n";
				BufferedReader processOut= new BufferedReader(new InputStreamReader(process.getInputStream()));
				while ( (line= processOut.readLine()) != null) {
					neverClaim+= line + "\n";
				}
				process.waitFor();

				// write pmlltlFile
				FileWriter pmlltlFileWriter= new FileWriter(pmlltlFile);
				pmlltlFileWriter.write(neverClaim);
				pmlltlFileWriter.flush();
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Never claim has been created.	 \n" + next;

				// verify the model
				String claimINpmlltlspin= commandOptions.get("claimINpmlltl.spin");
				if (claimINpmlltlspin == null || claimINpmlltlspin.equals("")) {
					call= spinLocation + " -a -N " + pmlltlFile.getAbsolutePath() + " " + pmlFile.getAbsolutePath();
				} else {
					call= spinLocation + " " + claimINpmlltlspin + " " + pmlltlFile.getAbsolutePath() + " " + pmlFile.getAbsolutePath();
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Verifying model (" + call + ")\n" + next;
				process= Runtime.getRuntime().exec(call);
				log= writeOutputToLog(process, log);
				process.waitFor();
				if (process.exitValue() != 0)
					throw new Exception("An error occured while running Spin.\n");

				// compile pan.c with gcc
				String calimINpmlltlgcc= commandOptions.get("claimINpmlltl.gcc");
				if (calimINpmlltlgcc == null || calimINpmlltlgcc.equals("")) {
					call= gccLocation + " -DXUSAFE -DNOREDUCE -DCOLLAPSE -o pan pan.c";
				} else {
					call= gccLocation + " " + calimINpmlltlgcc + " -o pan pan.c";
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Compiling (" + call + ")\n" + next;
				process= Runtime.getRuntime().exec(call);
				log= writeOutputToLog(process, log);
				process.waitFor();

				// execute pan.exe
				String pan= new File("pan").getAbsolutePath();
				String claimINpmlltlpan= commandOptions.get("claimINpmlltl.pan");
				if (claimINpmlltlpan != null && claimINpmlltlpan.equals("")) {
					call= pan;
				} else {
					call= pan + " " + claimINpmlltlpan;
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Executing (" + call + ")\n" + next;
				process= Runtime.getRuntime().exec(call);
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Output from executing pan: \n" + next;
				log= writeOutputToLog(process, log);
				process.waitFor();

			} catch (Exception e2) {
				// instantly return error message
				propertyCheck= "ERROR: An Exception occured while processing your request. See log for details. \n";
				log+= "[ERROR: checkPromelaFile] " + e2.getMessage();
				for (int i= 0; i < e2.getStackTrace().length; i++) {
					log+= "[JAVA: checkPromelaFile] " + e2.getStackTrace()[i].toString() + "\n";
				}
				result[PROPERTYCHECK]= propertyCheck;
				result[TRACE]= trace;
				result[LOG]= log;
				return result;
			}
		} else {
			try {
				log+= next + "[sf_ID_LOG: checkPromelaFile] "
						+ "Case: Never claim is NOT included in pml string and a pml.ltl string is empty or null. \n" + next;

				// verify the model
				if (commandOptions.get("noNeverClaim.spin").equals("")) {
					call= spinLocation + " -a " + pmlFile.getAbsolutePath();
				} else {
					call= spinLocation + " " + commandOptions.get("noNeverClaim.spin") + " " + pmlFile.getAbsolutePath();
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Verifying model (" + call + ")\n" + next;
				Process process= Runtime.getRuntime().exec(call);
				log= writeOutputToLog(process, log);
				process.waitFor();
				if (process.exitValue() != 0)
					throw new Exception("An error occured while running Spin.\n");

				// compile pan.c with gcc
				if (commandOptions.get("noNeverClaim.gcc").equals("")) {
					call= gccLocation + " -DXUSAFE -DNOCLAIM -DSAFETY -DCOLLAPSE -DREACH -o pan pan.c";
				} else {
					call= gccLocation + " " + commandOptions.get("noNeverClaim.gcc") + " -o pan pan.c";
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Compiling (" + call + ")\n" + next;
				process= Runtime.getRuntime().exec(call);
				log= writeOutputToLog(process, log);
				process.waitFor();

				// execute pan.exe
				String pan= new File("pan").getAbsolutePath();
				if (commandOptions.get("noNeverClaim.pan").equals("")) {
					call= pan + " -E -m1000 -i";
				} else {
					call= pan + " " + commandOptions.get("noNeverClaim.pan");
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Executing (" + call + ")\n" + next;
				process= Runtime.getRuntime().exec(call);
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Output from executing pan: \n" + next;
				log= writeOutputToLog(process, log);
				process.waitFor();

			} catch (Exception e2) {
				// instantly return error message
				propertyCheck= "ERROR: An Exception occured while processing your request. See log for details. \n";
				log+= "[ERROR: checkPromelaFile] " + e2.getMessage();
				for (int i= 0; i < e2.getStackTrace().length; i++) {
					log+= "[JAVA: checkPromelaFile] " + e2.getStackTrace()[i].toString() + "\n";
				}
				result[PROPERTYCHECK]= propertyCheck;
				result[TRACE]= trace;
				result[LOG]= log;
				return result;
			}


		}
		// analyze the result
		if (log.indexOf("pan: claim violated") > -1 || log.indexOf("pan: assertion violated") > -1 || log.indexOf("pan: acceptance cycle") > -1) {
			if (log.indexOf("pan: claim violated") > -1 || log.indexOf("pan: acceptance cycle") > -1) {
				propertyCheck= "Property is satisfied";
				log+= "[sf_ID_LOG: checkPromelaFile] " + "Property is satisfied! \n";
			} else {
				propertyCheck= "Property is NOT satisfied";
				log+= "[sf_ID_LOG: checkPromelaFile] " + "Property is NOT satisfied! \n";
			}

			try {
				// create a trail
				String createTrailspin= commandOptions.get("createTrail.spin");
				String trailOptions= (createTrailspin == null || createTrailspin.equals("")) ? "-l -g -w -v -t" : createTrailspin;
				if (!pmlltlfilecontent.equals("")) {
					call= spinLocation + " -N " + pmlltlFile.getAbsolutePath() + " " + trailOptions + " " + pmlFile.getAbsolutePath();
				} else {
					call= spinLocation + " " + trailOptions + " " + pmlFile.getAbsolutePath();
				}
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Creating trail (" + call + ")\n" + next;
				Process process= Runtime.getRuntime().exec(call);
				// capture output and write it to the tlFile, if no errors
				// occurred
				BufferedReader processOut= new BufferedReader(new InputStreamReader(process.getInputStream()));
				String line;
				while ( (line= processOut.readLine()) != null) {
					trace+= line + "\n";
				}
				process.waitFor();
				log+= next + "[sf_ID_LOG: checkPromelaFile] " + "Trail has been created.  \n" + next;
			} catch (Exception e2) {
				// instantly return error message
				propertyCheck= "ERROR: An Exception occured while processing your request. See log for details. \n";
				log+= "[ERROR: checkPromelaFile] " + e2.getMessage();
				for (int i= 0; i < e2.getStackTrace().length; i++) {
					log+= "[JAVA: checkPromelaFile] " + e2.getStackTrace()[i].toString() + "\n";
				}
				result[PROPERTYCHECK]= propertyCheck;
				result[TRACE]= trace;
				result[LOG]= log;
				return result;
			}

		} else {
			propertyCheck= "Property is NOT satisfied";
			log+= "[sf_ID_LOG: checkPromelaFile] " + "Property is NOT satisfied! \n";
		}
		// delete the fileNumber from the list of fileNumbers
		fileNumbers.remove(new Integer(fileNumber));
		// create the result String array
		result[PROPERTYCHECK]= propertyCheck;
		result[TRACE]= trace;
		result[LOG]= log;
		return result;
	}

	private String getStringRepresentation(Hashtable<String, String> hashtable) {
		String stringRepresentation= "";
		for (String key : hashtable.keySet()) {
			stringRepresentation+= "[sf_ID_LOG: checkPromelaFile] " + key + "= " + hashtable.get(key) + "\n";
		}
		return stringRepresentation;
	}

	private Hashtable<String, String> parseCommandOptions(String[] options) {
		Hashtable<String, String> commandOptions= new Hashtable<String, String>();
		for (int i= 0; i < options.length; i++) {
			String name= options[i].substring(0, options[i].indexOf("="));
			String value= options[i].substring(options[i].indexOf("=") + 1);
			if (value == null)
				value= "";
			commandOptions.put(name, value);
		}
		return commandOptions;
	}

	/**
	 * write the error message of a process to a log string
	 * 
	 * @param process - the process Object from which to retrieve the
	 *            errorStream
	 * @param log - the log file Object to write the error message to
	 * @throws Exception
	 */
	private String writeOutputToLog(Process process, String log) throws Exception {
		String line;
		BufferedReader processLog= new BufferedReader(new InputStreamReader(process.getInputStream()));
		while ( (line= processLog.readLine()) != null) {
			log+= "[sf_ID_LOG: checkPromelaFile] " + line + "\n";
		}
		BufferedReader processError= new BufferedReader(new InputStreamReader(process.getErrorStream()));
		while ( (line= processError.readLine()) != null) {
			log+= "[sf_ID_LOG: checkPromelaFile] " + line + "\n";
		}
		return log;
	}
}
